from just_playback import Playback


class InvalidFileType(Exception):
    pass


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class MusicPlayer(metaclass=Singleton):
    def __init__(self, default_volume: float = 0.5):
        self.playback = Playback()
        self.playback.set_volume(default_volume)
        self._loop_at_end = False

    def unload_song(self) -> None:
        self.playback.stop()

    def load_song(self, path: str) -> None:
        self.playback.load_file(str(path))

    def play_song(self) -> None:
        self.playback.play()

    def resume_song(self) -> None:
        self.playback.resume()

    def pause_song(self) -> None:
        self.playback.pause()

    def play_pause(self) -> None:
        if self.playback.playing:
            self.pause_song()
        else:
            self.resume_song()

    @property
    def loop_at_end(self) -> bool:
        return self._loop_at_end

    @loop_at_end.setter
    def loop_at_end(self, value: bool) -> None:
        self._loop_at_end = value
        self.playback.loop_at_end(value)

    @property
    def volume(self) -> float:
        return self.playback.volume

    @volume.setter
    def volume(self, volume: float):
        self.playback.set_volume(volume)

    @property
    def playing(self) -> bool:
        return self.playback.playing

    @property
    def position(self) -> float:
        return self.playback.curr_pos

    @position.setter
    def position(self, value: float) -> None:
        self.playback.seek(value)

    @property
    def active(self) -> bool:
        return self.playback.active

    @property
    def song_length(self) -> float:
        return self.playback.duration



from pytubefix import YouTube, Stream
from pydub import AudioSegment
from pathlib import Path
from typing import Callable
import music_tag
from .ytmusic import SongData

from PIL import Image
import io


def image_to_byte(image: Image) -> bytes:
    # BytesIO is a file-like buffer stored in memory
    imgByteArr = io.BytesIO()
    # image.save expects a file-like as a argument
    image.save(imgByteArr, format=image.format)
    # Turn the BytesIO object back into a bytes object
    imgByteArr = imgByteArr.getvalue()
    return imgByteArr


def _download_from_yt(
    song: SongData,
    download_path: str,
    callback: None | Callable[[Stream, bytes, int], None] = None,
) -> str | None:
    """Download a song from a song

    Args:
        song (SearchSongResult): The song to download

    Returns:
        path (str): The path of the downloaded file or None if the download failed
    """
    try:
        yt = YouTube(
            f"https://www.youtube.com/watch?v={song.videoId}",
            on_progress_callback=callback,
        )
        return yt.streams.get_audio_only().download(
            output_path=download_path,
            filename=f"{song.videoId}.m4a",
        )
    except Exception as e:
        print(e)
        return None


def _convert_to_mp3(path: str) -> str:
    """Convert an audio file to mp3

    Args:
        path (str): path to the downloanded AUDIO file

    Returns:
        str: path to the mp3 file
    """

    if not isinstance(path, str):
        raise TypeError(f"path must be a string, not {type(path)}")
    extention = Path(path).suffix
    new_path = path.replace(extention, ".mp3")
    audio = AudioSegment.from_file(path)
    audio.export(new_path, format="mp3")
    return new_path


def _delete_file(path: str) -> None:
    """Delete a file

    Args:
        path (str): path to the file
    """
    Path(path).unlink(missing_ok=True)


class Downloader:
    def __init__(self, download_path: str) -> None:
        self.download_path = download_path

    def download(self, song: SongData) -> str | None:
        """Download a song from a song object and return the path of the downloaded file.
        If the file already exists, it will not be downloaded again and the path will be returned.

        Args:
            song (Song): The song to download

        Returns:
            path (str): The path of the downloaded file or None if the download failed
        """
        self.song = song

        song_path = Path(f"{self.download_path}/{song.videoId}.mp3")
        if song_path.exists():
            return str(song_path)

        yt_path = _download_from_yt(song, self.download_path, self.on_progress)

        if yt_path is None:
            return None

        converted_path = _convert_to_mp3(yt_path)

        file_path = music_tag.load_file(converted_path)
        file_path["title"] = song.title
        file_path["artist"] = [artist for artist in song.artist]
        file_path["artwork"] = image_to_byte(song.thumbnail)
        file_path["album"] = song.album
        file_path.save()

        _delete_file(yt_path)

        return str(converted_path)

    def on_progress(self, stream: Stream, chunk: bytes, bytes_remaining: int) -> None:
        filesize = stream.filesize
        bytes_received = filesize - bytes_remaining
        percent = (bytes_received / filesize) * 100
        print(f"Downloaded {percent:.2f}% of {stream.title}")



import pytest
from pathlib import Path
from api.downloader import Downloader, _convert_to_mp3, _download_from_yt, _delete_file
from util_test import Cleaner


class MockSongData:
    title = "Test Song"
    duration = "3:30"
    videoId = "C0DPdy98e4c"

    def get_formatted_artists(self) -> str:
        return "Test Artist"


BASE_PATH = Path(__file__).parent / "tmp"


@pytest.fixture
def downloader():
    return Downloader(download_path=BASE_PATH)


def test_downloader_downloading_file(downloader):
    with Cleaner(BASE_PATH):
        song = MockSongData()
        song_path = (
            Path(BASE_PATH) / f"{song.title} - {song.get_formatted_artists()}.mp3"
        )
        result = downloader.download(song)
        assert result == str(song_path)
        assert Path(result).exists()


def test_delete_file():
    with Cleaner(BASE_PATH):
        path = BASE_PATH / "Test Song - Test Artist.m4a"
        Path(path).touch()  # Create the file to simulate it exists

        _delete_file(path)
        assert not Path(path).exists()


def test_song_already_downloaded(downloader):
    with Cleaner(BASE_PATH):
        song = MockSongData()
        song_path = (
            Path(BASE_PATH) / f"{song.title} - {song.get_formatted_artists()}.mp3"
        )
        song_path.touch()  # Create the file to simulate it exists
        result = downloader.download(song)
        assert result == str(song_path)
        assert Path(result).exists()


def test_download_failed_return_none():
    with Cleaner(BASE_PATH):
        song = MockSongData()
        song.videoId = "invalid"
        result = _download_from_yt(song, BASE_PATH)
        assert result is None


def test_downloader_failed_return_none(downloader):
    with Cleaner(BASE_PATH):
        song = MockSongData()
        song.videoId = "invalid"
        result = downloader.download(song)
        assert result is None


def test_download_from_yt():
    with Cleaner(BASE_PATH):
        song = MockSongData()
        song_path = _download_from_yt(song, BASE_PATH)
        assert Path(song_path).exists()
        assert Path(song_path).suffix == ".m4a"
        assert (
            Path(song_path).name == f"{song.title} - {song.get_formatted_artists()}.m4a"
        )


def test_convert_to_mp3():
    with Cleaner(BASE_PATH):
        song = MockSongData()
        song_path = _download_from_yt(song, BASE_PATH)
        mp3_path = _convert_to_mp3(song_path)
        assert Path(mp3_path).exists()
        assert Path(mp3_path).suffix == ".mp3"
        assert (
            Path(mp3_path).name == f"{song.title} - {song.get_formatted_artists()}.mp3"
        )


def test_convert_to_mp3_type_error():
    with pytest.raises(TypeError):
        _convert_to_mp3(90)



from typing import Protocol

from setting import SettingManager

setting = SettingManager()
if setting.os == "win32":
    pass
else:
    from mpris_server import EventAdapter
    from mpris_server import Server as ServerMpris
    from api.mpris.mpris import HAdapter


class MediaControl(Protocol):
    def init(self) -> None: ...
    def on_playback(self) -> None: ...
    def on_playpause(self) -> None: ...
    def on_volume(self) -> None: ...


class Server(Protocol):
    root: str
    player: str


class MediaControlWin32(MediaControl):
    def __init__(self) -> None: ...

    def init(self, player) -> None: ...

    def on_playback(self) -> None: ...

    def on_playpause(self) -> None: ...

    def on_volume(self) -> None: ...


class MediaControlMPRIS(MediaControl):
    def __init__(self) -> None:
        self.adapter = HAdapter()
        self.mpris = ServerMpris(name="PyMusicTerm", adapter=self.adapter)
        self.event = EventAdapter(root=self.mpris.root, player=self.mpris.player)

    def init(self, player):
        self.adapter.setup(player)
        self.mpris.loop(background=True)

    def on_playback(self):
        return self.event.on_playback()

    def on_playpause(self):
        return self.event.on_playpause()

    def on_volume(self):
        return self.event.on_volume()



from datetime import timedelta
from api.protocols import SongData
from player.player import PyMusicTermPlayer
from textual.app import App, ComposeResult
from textual.widget import Widget
from textual.widgets import (
    Input,
    Button,
    TabbedContent,
    Label,
    TabPane,
    Rule,
    Select,
    MarkdownViewer,
    ProgressBar,
    ListView,
    ListItem,
)
from textual.containers import Horizontal, Vertical, Center
from textual import on
from pathlib import Path
from textual import work
from textual.worker import get_current_worker
from loguru import logger
from textual_image.widget import Image as WidgetImage
from player.util import format_time
from setting import SettingManager, rename_console
import requests_cache
# TODO : rendre plus maintenable le code existnant en refactorisant
# TODO : renomer les fonctions pour qu'elles soient plus explicites, et pareil pour les id et classes des widgets
# TODO : ajouter des tests unitaires
# TODO : améliorer le queuing systeme
# TODO : ajouter des raccourcis clavier modifiables
# TODO : AJOUTER Type validation (raise error) --- 50%
# TODO : ajouter des logs messages


class PyMusicTerm(App):
    BINDINGS = [
        ("q", "seek_back", "Seek backward"),
        ("s", "play", "Play/Pause"),
        ("d", "seek_forward", "Seek forward"),
        ("r", "shuffle", "Shuffle"),
        ("l", "loop", "Loop at the end"),
        ("a", "previous", "Previous song"),
        ("e", "next", "Next song"),
        ("&", "return_on_search_tab", "Go to the search tab"),
        ("é", "return_on_playlist_tab", "Go to the playlist tab"),
        ('"', "return_on_lyrics_tab", "Go to the lyrics tab"),
        ("j", "volume(-0.01)", "Volume down"),
        ("k", "volume(0.01)", "Volume up"),
    ]

    def __init__(self, setting: SettingManager) -> None:
        super().__init__(css_path="pymusicterm.tcss", watch_css=True)
        self.setting = setting
        rename_console("PyMusicTerm")

        logger.remove()
        logger.add(
            Path(self.setting.log_dir + "/pymusicterm.log"),
            format="{time} {level} {message}",
            level="INFO",
            rotation="500 MB",
        )
        self.timer: Widget | None = None

        if self.setting.os == "win32":
            from player.media_control import MediaControlWin32 as MediaControl
        else:
            from player.media_control import MediaControlMPRIS as MediaControl
        requests_cache.install_cache(
            f"{self.setting.cache_dir}/cache", expire_after=timedelta(hours=1)
        )

        self.media_control = MediaControl()
        self.player = PyMusicTermPlayer(self.setting, self.media_control)
        self.media_control.init(self.player)

    def compose(self) -> ComposeResult:
        with TabbedContent(classes="search_tabs", id="tabbed_content"):
            with TabPane("Search", id="search"):
                with Vertical():
                    yield Input(placeholder="Search for a song", id="search_input")
                    yield Select(
                        id="search_sort",
                        options=[
                            ("Songs (default: Stream Youtube Music)", "songs"),
                            ("Video (Stream Youtube Video)", "videos"),
                        ],
                        allow_blank=False,
                    )
                    yield ListView(id="search_results")
            with TabPane("Playlist", id="playlist"):
                with Vertical():
                    yield Input(placeholder="Search for a song", id="playlist_input")
                    yield ListView(id="playlist_results")
            with TabPane("Lyrics", id="lyrics"):
                with Vertical():
                    yield Input(placeholder="Search for a song", id="lyrics_input")
                    yield MarkdownViewer(
                        markdown="No lyrics now",
                        id="lyrics_results",
                        show_table_of_contents=False,
                    )
        yield Rule()
        with Vertical(classes="info_controls"):
            with Center():
                yield Label("Unknown Title", id="label_current_song_title")
            with Center():
                yield Label("Unknown Artist", id="label_current_song_artist")
        with Horizontal(classes="status_controls"):
            yield Label(
                "--:--", id="label_current_song_position", classes="control_label"
            )
            yield ProgressBar(
                total=100,
                show_eta=False,
                show_percentage=False,
                id="player_status",
            )
            yield Label("--:--", id="label_song_length", classes="control_label")
        with Horizontal(classes="player_controls"):
            yield Button("󰼨", id="previous")
            yield Button("󰐊", id="play_pause")
            yield Button("󰼧", id="next")
        with Horizontal(classes="player_controls"):
            yield Button("󰒟", id="shuffle")
            yield Button("󰛤", id="loop")

    @on(TabbedContent.TabActivated)
    def action_select_playlist_tab(self, event: TabbedContent.TabActivated) -> None:
        """Action when the tab is activated. If the tab is playlist, clear the options and add the songs to the playlist.

        Args:
            event (TabbedContent.TabActivated): The event that triggered the action
        """
        playlist_results: ListView = self.query_one("#playlist_results")
        if event.tab.id.endswith("playlist"):
            children_id = [
                child.id.removeprefix("id-") for child in playlist_results.children
            ]
            for song in self.player.list_of_downloaded_songs:
                if song.videoId not in children_id:
                    playlist_results.append(self._create_song_item(song))

    def update_time(self) -> None:
        """Update the time label of the player, and update the player"""
        progress_bar: ProgressBar = self.query_one("#player_status")
        label_current_song_position: Label = self.query_one(
            "#label_current_song_position"
        )
        label_song_length: Label = self.query_one("#label_song_length")
        length_float = self.player.song_length
        current_float = self.player.position
        label_current_song_position.update(format_time(current_float))
        label_song_length.update(format_time(length_float))
        if self.player.playing:
            label_current_song_title: Label = self.query_one(
                "#label_current_song_title"
            )
            label_current_song_artist: Label = self.query_one(
                "#label_current_song_artist"
            )
            label_current_song_title.update(
                self.player.list_of_downloaded_songs[
                    self.player.current_song_index
                ].title
            )
            label_current_song_artist.update(
                self.player.list_of_downloaded_songs[
                    self.player.current_song_index
                ].get_formatted_artists()
            )
        try:
            percentage = current_float / length_float
        except ZeroDivisionError:
            percentage = 0.01
        progress_bar.update(
            progress=percentage * 100,
        )
        self.player.check_if_song_ended()

    def action_return_on_search_tab(self):
        """Set the search tab as the active tab"""
        tab: TabbedContent = self.query_one("#tabbed_content")
        tab.active = "search"

    def action_return_on_playlist_tab(self):
        """Set the playlist tab as the active tab"""
        tab: TabbedContent = self.query_one("#tabbed_content")
        tab.active = "playlist"

    def action_return_on_lyrics_tab(self):
        """Set the lyrics tab as the active tab"""
        tab: TabbedContent = self.query_one("#tabbed_content")
        tab.active = "lyrics"

    @on(ListView.Selected, "#playlist_results")
    def select_playlist_result(self, event: ListView.Selected) -> None:
        """Select a song from the playlist results and play it"""
        id = event.item.id.removeprefix("id-")
        self.play_from_id(id)

    def play_from_id(self, ids: str) -> None:
        """Play a song from the playlist results and play it

        Args:
            id (str): The id of the song to play
        """
        for i, song in enumerate(self.player.list_of_downloaded_songs):
            if song.videoId == ids:
                self.player.play_from_list(i)
                self.toggle_button()

    @on(ListView.Selected, "#search_results")
    def select_result(self, event: ListView.Selected) -> None:
        """Select a song from the search results and play it"""
        video_id = str(event.item.id).removeprefix("id-")
        self.player.play_from_ytb(video_id)
        self.toggle_button()

    @on(Button.Pressed, "#previous")
    def action_previous(self) -> None:
        """Play the previous song"""
        self.toggle_button()
        playlist_results: ListView = self.query_one("#playlist_results")
        playlist_results.index = self.player.previous()

    @on(Button.Pressed, "#play_pause")
    def action_play(self) -> None:
        """Play or pause the song"""
        if self.player.playing:
            self.player.pause_song()
        else:
            self.player.resume_song()
        self.toggle_button()

    def toggle_button(self) -> None:
        """Toggle the play button label and start the timer if it's not running"""
        button: Button = self.query_one("#play_pause")
        if self.player.playing:
            button.label = "󰏤"
        else:
            button.label = "󰐊"
        if self.timer is None:
            self.timer = self.set_interval(0.1, self.update_time, name="update_time")

    @on(Button.Pressed, "#next")
    def action_next(self) -> None:
        """Play the next song"""
        self.toggle_button()
        playlist_results: ListView = self.query_one("#playlist_results")
        playlist_results.index = self.player.next()

    @on(Button.Pressed, "#shuffle")
    def action_shuffle(self) -> None:
        """Shuffle the list of downloaded songs"""
        self.player.suffle()
        playlist_results: ListView = self.query_one("#playlist_results")
        playlist_results.clear()
        for i, song in enumerate(self.player.list_of_downloaded_songs):
            ListItem(
                Label(f"{i + 1}. {song.title} - {song.get_formatted_artists()}"),
            )

    @on(Button.Pressed, "#loop")
    def action_loop(self) -> None:
        """Toggle the loop button"""
        loop_button: Button = self.query_one("#loop")
        is_looping = self.player.loop_at_end()
        if is_looping:
            loop_button.variant = "success"
        else:
            loop_button.variant = "default"

    @on(Input.Submitted, "#search_input")
    def search(self) -> None:
        """Search for a song on YTMusic and display the results"""
        search_results: ListView = self.query_one("#search_results")
        search_results.clear()
        search_input: Input = self.query_one("#search_input")
        if search_input.value == "":
            return
        search_results.loading = True
        self.search_ytb_thread(search_input.value)

    @work(exclusive=True, thread=True)
    def search_ytb_thread(self, query: str) -> None:
        """Start a thread to search for a song on YTMusic

        Args:
            query (str): The search query
        """
        worker = get_current_worker()
        filter: Select = self.query_one("#search_sort")
        results = self.player.query(query, filter.value)
        if not worker.is_cancelled:
            self.call_from_thread(self.update_search_results, results)

    def update_search_results(self, results: list[SongData]) -> None:
        """Update the search results

        Args:
            results (list): The list of results from youtube music
        """
        search_results: ListView = self.query_one("#search_results")
        search_results.clear()
        for result in results:
            search_results.append(self._create_song_item(result))
        search_results.loading = False

    def _create_song_item(self, song: SongData) -> ListItem:
        """Create a song item for the search results

        Args:
            song (SearchResult): The song to create the item for

        Returns:
            ListItem: The song item
        """

        return ListItem(
            Horizontal(
                WidgetImage(song.thumbnail, classes="image"),
                Vertical(
                    Label(
                        song.title,
                        classes="title",
                        markup=False,
                    ),
                    Label(
                        song.get_formatted_artists(),
                        markup=False,
                        classes="artist",
                    ),
                ),
                Label(
                    song.album,
                    markup=False,
                    classes="album",
                ),
                Label(
                    song.duration,
                    markup=False,
                    classes="length",
                ),
            ),
            id=f"id-{song.videoId}",
            classes="song_item",
        )

    def action_seek_back(self) -> None:
        """Seek backward 10 seconds"""
        self.player.seek(-10)

    def action_seek_forward(self) -> None:
        """Seek forward 10 seconds"""
        self.player.seek(10)

    def action_volume(self, volume: float) -> None:
        """Increase the volume"""
        self.player.volume(volume)
        self.notify(f"Volume changed to {self.player.music_player.volume:.2f}")


@logger.catch
def main():
    setting = SettingManager()
    app = PyMusicTerm(setting)
    app.run()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from PIL import ImageFile


@dataclass
class SongData:
    title: str
    artist: list[str]
    duration: str
    videoId: str
    thumbnail: ImageFile
    album: str
    path: None | str = None

    def get_formatted_artists(self) -> str:
        """Get the formatted artists of the song"""
        return ", ".join([artist for artist in self.artist])



import requests
import ytmusicapi
from dataclasses import dataclass
import ytmusicapi.exceptions
from PIL import Image

from api.protocols import SongData


@dataclass
class LyricsResult:
    lyrics: str
    source: str


class YTMusic:
    def __init__(self):
        self.client = ytmusicapi.YTMusic()

    def search(self, query: str, filter: str = "songs") -> list[SongData]:
        """Search for a song on YTMusic
        Args:
            query (str): The query to search for
            filter (str, optional): The filter to use. Defaults to "songs".

        Raises:
            TypeError: If query or filter is not a string

        Returns:
            list[Song]: The list of songs found
        """
        if not isinstance(query, str):
            raise TypeError(f"query must be a string, not {type(query)}")
        if not isinstance(filter, str):
            raise TypeError(f"filter must be a string, not {type(filter)}")

        results = self.client.search(query, filter)
        r = []
        for result in results:
            title: str = result.get("title", "Unknown")
            artist: list[str] = [artist["name"] for artist in result.get("artists", [])]
            duration: str = result.get("duration", "Unknown")
            videoId: str = result.get("videoId", "Unknown")
            thumbnail = Image.open(
                requests.get(result["thumbnails"][0]["url"], stream=True).raw
            )
            x = result.get("album", None)
            if x:
                album = x.get("name")
            else:
                album = "Unknown"
            r.append(
                SongData(
                    title=title,
                    artist=artist,
                    duration=duration,
                    videoId=videoId,
                    thumbnail=thumbnail,
                    album=album,
                )
            )
        return r



import subprocess
from setting import Setting

if __name__ == "__main__":
    setting = Setting()
    subprocess.run(
        [
            "c:/Users/Zach/PycharmProjects/PyMusicTerm/.venv/Scripts/python.exe",
            "c:/Users/Zach/PycharmProjects/PyMusicTerm/src/main.py",
        ],
        shell=True,
    )



from dataclasses import dataclass
from pathlib import Path
from msgspec import toml
import os
import sys

HOME = Path.home()
APP_DIR = Path(HOME / ".pymusicterm")
MUSIC_DIR = Path(APP_DIR / "musics")
SETTING_FILE = Path(APP_DIR / "setting.toml")
PLAYLIST_DIR = Path(APP_DIR / "playlists")
LYRICS_DIR = Path(APP_DIR / "lyrics")
LOG_DIR = Path(APP_DIR / "logs")
CACHE_DIR = Path(APP_DIR / "cache")


def fetch_files_from_folder(folder_path: str, ending: str = "mp3") -> list[str | None]:
    """Fetch all the files from a folder
    Args:
        folder_path (str): The path of the folder
        ending (str): The ending of the files to fetch. Defaults to "mp3"
    Returns:
        list[str]: The list of file found
    """
    if not isinstance(folder_path, str):
        raise TypeError(f"folder_path must be a string, not {type(folder_path)}")
    if not isinstance(ending, str):
        raise TypeError(f"ending must be a string, not {type(ending)}")
    return [str(file) for file in Path(folder_path).glob(f"*.{ending}")]


def resource_path(relative_path):
    """Get absolute path to resource, works for dev and for PyInstaller"""
    base_path = getattr(sys, "_MEIPASS", os.path.dirname(os.path.abspath(__file__)))
    return os.path.join(base_path, relative_path)


@dataclass
class Setting:
    """All the settings of the app"""

    volume: float = 1.0
    loop: bool = False
    os: str = sys.platform
    app_dir: str = str(APP_DIR)
    music_dir: str = str(MUSIC_DIR)
    setting_file: str = str(SETTING_FILE)
    playlist_dir: str = str(PLAYLIST_DIR)
    lyrics_dir: str = str(LYRICS_DIR)
    log_dir: str = str(LOG_DIR)
    cache_dir: str = str(CACHE_DIR)


class SettingManager:
    """Manages the settings of the app."""

    def __init__(self):
        self._setting = None
        self.check_and_create_paths()
        self._setting = self.load_setting()

    @property
    def os(self) -> str:
        return self._setting.os

    @os.setter
    def os(self, value: str) -> None:
        self._setting.os = value
        self.save_setting()

    @property
    def volume(self) -> float:
        return self._setting.volume

    @volume.setter
    def volume(self, value: float) -> None:
        self._setting.volume = round(value, 3)
        self.save_setting()

    @property
    def loop(self) -> bool:
        return self._setting.loop

    @loop.setter
    def loop(self, value: bool) -> None:
        self._setting.loop = value
        self.save_setting()

    @property
    def app_dir(self) -> str:
        return self._setting.app_dir

    @property
    def music_dir(self) -> str:
        return self._setting.music_dir

    @property
    def setting_file(self) -> str:
        return self._setting.setting_file

    @property
    def playlist_dir(self) -> str:
        return self._setting.playlist_dir

    @property
    def lyrics_dir(self) -> str:
        return self._setting.lyrics_dir

    @property
    def log_dir(self) -> str:
        return self._setting.log_dir

    @property
    def cache_dir(self) -> str:
        return self._setting.cache_dir

    def load_setting(self) -> Setting:
        """Load settings from the setting.toml file."""
        if not SETTING_FILE.exists():
            self._setting = Setting()
            self.save_setting()

        try:
            with open(SETTING_FILE, "rb") as f:
                return toml.decode(f.read(), type=Setting)
        except Exception as e:
            print(f"Error loading settings: {e}")
            return Setting()

    def save_setting(self) -> None:
        """Save the current settings to the setting.toml file."""
        try:
            with open(SETTING_FILE, "wb") as f:
                encoded = toml.encode(self._setting)
                f.write(encoded)
        except Exception as e:
            print(f"Error saving settings: {e}")

    def check_and_create_paths(self) -> None:
        """Check and create necessary directories and files."""
        APP_DIR.mkdir(exist_ok=True)
        MUSIC_DIR.mkdir(exist_ok=True)
        PLAYLIST_DIR.mkdir(exist_ok=True)
        LYRICS_DIR.mkdir(exist_ok=True)
        LOG_DIR.mkdir(exist_ok=True)
        CACHE_DIR.mkdir(exist_ok=True)


@dataclass
class KeyBinding:
    """All the keybindings of the app"""

    volume_up: str = "k"
    volume_down: str = "j"
    seek_back: str = "q"
    seek_forward: str = "d"
    play_pause: str = "s"


def rename_console(name: str, platform: str = sys.platform) -> None:
    """Rename the console

    Args:
        name (str): The new name of the console (for linux and windows)
    """
    if not isinstance(name, str):
        raise TypeError(f"name must be a string, not {type(name)}")
    if platform not in ("win32", "linux"):
        raise ValueError(f"platform must be 'win32' or 'linux', not {platform}")
    if platform == "win32":
        os.system(f"title {name}")
    else:
        print(f"\33]0;{name}\a", end="", flush=True)


'''def have_internet() -> bool:
    """Check if the user has internet connection"""
    conn = httplib.HTTPSConnection("8.8.8.8", timeout=5)
    try:
        with conn.request("HEAD", "/"):
            return True
    except Exception:
        return False
'''



import sys
from unittest.mock import MagicMock
from setting import (
    APP_DIR,
    PLAYLIST_DIR,
    LYRICS_DIR,
    LOG_DIR,
    MUSIC_DIR,
    SETTING_FILE,
    SettingManager,
    rename_console,
    fetch_files_from_folder,
    Setting,
)
import pytest
from pathlib import Path
import uuid
from util_test import Cleaner

BASE_PATH = Path(__file__).parent / "tmp"


def test_rename_console_name_error():
    with pytest.raises(TypeError):
        rename_console(90)


def test_rename_console_platform_error():
    with pytest.raises(ValueError):
        rename_console("name", 90)


def test_rename_console():
    rename_console("name", "win32")
    rename_console("name", "linux")
    assert True


def test_fetch_files_from_folder_path_error():
    with pytest.raises(TypeError):
        fetch_files_from_folder(90)


def test_fetch_files_from_folder_ending_error():
    with pytest.raises(TypeError):
        fetch_files_from_folder("path", 90)


def test_fetch_files_from_folder():
    with Cleaner(BASE_PATH):
        file_names = [str(Path(BASE_PATH) / f"{uuid.uuid4()}.txt") for _ in range(10)]
        for file_name in file_names:
            Path(file_name).touch()

        files = fetch_files_from_folder(str(BASE_PATH), "txt")
        assert len(files) == 10
        assert all([file in file_names for file in files])


def test_setting_initialization():
    setting = Setting()
    assert setting.volume == 1.0
    assert setting.loop is False
    assert setting.os == sys.platform
    assert setting.app_dir == str(APP_DIR)
    assert setting.music_dir == str(MUSIC_DIR)
    assert setting.setting_file == str(SETTING_FILE)
    assert setting.playlist_dir == str(PLAYLIST_DIR)
    assert setting.lyrics_dir == str(LYRICS_DIR)
    assert setting.log_dir == str(LOG_DIR)


def test_setting_manager_load_save():
    setting_manager = SettingManager()
    setting_manager.volume = 0.5
    setting_manager.loop = True
    setting_manager.os = "test_os"

    loaded_setting = setting_manager.load_setting()
    assert loaded_setting.volume == 0.5
    assert loaded_setting.loop is True
    assert loaded_setting.os == "test_os"

    setting_manager.volume = 1.0
    setting_manager.loop = False
    setting_manager.os = sys.platform
    setting_manager.save_setting()

    loaded_setting = setting_manager.load_setting()
    assert loaded_setting.volume == 1.0
    assert loaded_setting.loop is False
    assert loaded_setting.os == sys.platform



import pytest
from unittest.mock import MagicMock
from main import PyMusicTerm, format_time
from setting import SettingManager
from textual.widgets import OptionList


def test_format_time():
    assert format_time(90) == "01:30"
    assert format_time(3600) == "1:00:00"
    assert format_time(3661) == "1:01:01"
    assert format_time(0) == "00:00"
    assert format_time(-100) == "00:00"
    with pytest.raises(TypeError):
        format_time("90")


@pytest.fixture
def app() -> PyMusicTerm:
    setting = SettingManager()
    setting.os = "win32"
    app = PyMusicTerm(setting)
    return app


def test_app_initialization(app: PyMusicTerm):
    assert app.setting is not None
    assert app.player is not None
    assert app.media_control is not None


@pytest.mark.asyncio
async def test_action_shuffle(app: PyMusicTerm):
    async with app.run_test() as pilot:
        await pilot.click("#playlist")
        playlist_results: OptionList = app.query_one("#playlist_results")
        previous_results = list(playlist_results.options)
        app.action_shuffle()
        assert previous_results != list(playlist_results.options)


@pytest.mark.asyncio
async def test_play_pause_button(app: PyMusicTerm):
    """Test pressing keys has the desired result."""
    async with app.run_test() as pilot:
        await pilot.click("#playlist")
        await pilot.click("#playlist_results")
        await pilot.click("#play_pause")
        assert app.player.playing is False
        await pilot.click("#play_pause")
        assert app.player.playing is False


def test_action_seek_back(app: PyMusicTerm):
    """Test if the seek back action works as expected."""
    app.player.seek = MagicMock()
    app.action_seek_back()
    app.player.seek.assert_called_once_with(-10)


def test_action_seek_forward(app: PyMusicTerm):
    """Test if the seek forward action works as expected."""
    app.player.seek = MagicMock()
    app.action_seek_forward()
    app.player.seek.assert_called_once_with(10)


def test_action_volume(app: PyMusicTerm):
    """test if the volume action works as expected."""
    app.player.volume = MagicMock()
    app.action_volume(0.1)
    app.player.volume.assert_called_once_with(0.1)



import pytest
from unittest.mock import MagicMock

import ytmusicapi
from api.ytmusic import YTMusic, SearchResult, SongArtist, LyricsResult
from api.lyrics import LyricsDownloader


@pytest.fixture
def lyrics_downloader():
    return MagicMock(spec=LyricsDownloader)


@pytest.fixture
def ytmusic(lyrics_downloader):
    return YTMusic(lyrics_downloader)


def test_search(ytmusic: YTMusic):
    ytmusic.client.search = MagicMock(
        return_value=[
            {
                "title": "Song Title",
                "artists": [{"name": "Artist Name", "id": "artist_id"}],
                "duration": 300,
                "videoId": "video_id",
            }
        ]
    )
    results = ytmusic.search("Song Title")
    assert len(results) == 1
    assert results[0].title == "Song Title"
    assert results[0].artist[0].name == "Artist Name"
    assert results[0].duration == 300
    assert results[0].videoId == "video_id"


def test_get_formated_artist(ytmusic):
    song = SearchResult(
        title="Song Title",
        artist=[
            SongArtist(name="Artist Name", id="artist_id"),
            SongArtist(name="Artist Name 2", id="artist_id2"),
        ],
        duration=300,
        videoId="video_id",
    )
    assert song.get_formatted_artists() == "Artist Name, Artist Name 2"


def test_search_invalid_filter(ytmusic):
    with pytest.raises(TypeError):
        ytmusic.search("query", 123)


def test_search_invalid_query(ytmusic):
    with pytest.raises(TypeError):
        ytmusic.search(123)


def test_get_lyrics(ytmusic: YTMusic):
    song = SearchResult(
        title="Song Title",
        artist=[SongArtist(name="Artist Name", id="artist_id")],
        duration=300,
        videoId="video_id",
    )
    ytmusic.client.get_watch_playlist = MagicMock(return_value={"lyrics": "lyrics_id"})
    ytmusic.client.get_lyrics = MagicMock(
        return_value={"lyrics": "Lyrics text", "source": "Source"}
    )
    ytmusic.lyrics_downloader.save = MagicMock(
        return_value=LyricsResult(lyrics="Lyrics text", source="Source")
    )

    lyrics_result: LyricsResult = ytmusic.get_lyrics(song)
    assert lyrics_result.lyrics == "Lyrics text"
    assert lyrics_result.source == "Source"


def test_get_lyrics_no_lyrics(ytmusic: YTMusic):
    song = SearchResult(
        title="Song Title",
        artist=[SongArtist(name="Artist Name", id="artist_id")],
        duration=300,
        videoId="video_id",
    )
    ytmusic.client.get_watch_playlist = MagicMock(return_value={"lyrics": "lyrics_id"})
    ytmusic.client.get_lyrics = MagicMock(
        side_effect=ytmusicapi.exceptions.YTMusicError
    )
    ytmusic.lyrics_downloader.save = MagicMock(
        return_value=LyricsResult(lyrics="None", source="None")
    )

    lyrics_result = ytmusic.get_lyrics(song)
    assert lyrics_result.lyrics == "None"
    assert lyrics_result.source == "None"



from typing import List, Protocol
from mpris_server.adapters import PlayState, MprisAdapter
from mpris_server.base import URI, MIME_TYPES
import sys
from mpris_server import MetadataObj, Track
from typing import override
from api.ytmusic import SongData


class PyMusicTermPlayer(Protocol):
    position: float
    current_song_index: int
    song_length: float
    playing: bool
    list_of_downloaded_songs: list[SongData]
    dict_of_lyrics: dict[str, str]

    def query(self, query: str, filter: str) -> list: ...
    def play_from_ytb(self, video_id: str) -> None: ...
    def play_from_list(self, id: int) -> None: ...
    def previous(self) -> None: ...
    def next(self) -> None: ...
    def seek_to(self, time: float) -> None: ...
    def seek(self, time: float = 10) -> None: ...
    def suffle(self) -> None: ...
    def loop_at_end(self) -> bool: ...
    def update(self) -> None: ...
    def stop(self) -> None: ...
    def pause_song(self) -> None: ...
    def resume_song(self) -> None: ...


class HAdapter(MprisAdapter):
    def __init__(self):
        super().__init__()
        self.player: PyMusicTermPlayer | None = None

    def setup(self, player):
        self.player = player

    @override
    def get_uri_schemes(self) -> List[str]:
        return URI

    @override
    def get_mime_types(self) -> List[str]:
        return MIME_TYPES

    @override
    def can_quit(self) -> bool:
        return True

    @override
    def quit(self):
        sys.exit()

    @override
    def get_current_position(self):
        return self.player.position

    @override
    def next(self):
        self.player.next()

    @override
    def previous(self):
        self.player.previous()

    @override
    def pause(self):
        self.player.pause_song()

    @override
    def resume(self):
        self.player.resume_song()

    @override
    def stop(self):
        self.player.stop()

    @override
    def play(self):
        self.player.playing

    @override
    def get_playstate(self) -> PlayState:
        if not self.player.playing:
            return PlayState.PAUSED
        else:
            return PlayState.PLAYING

    @override
    def seek(self, time, track_id=None):
        self.player.seek_to(time / 1000000)

    @override
    def is_repeating(self) -> bool:
        return False

    @override
    def is_playlist(self) -> bool:
        return self.can_go_next() or self.can_go_previous()

    @override
    def set_repeating(self, val: bool):
        self.player.loop_at_end()

    @override
    def set_loop_status(self, val: str):
        pass

    @override
    def get_rate(self) -> float:
        return 1.0

    @override
    def set_rate(self, val: float):
        pass

    @override
    def get_shuffle(self) -> bool:
        return False

    @override
    def set_shuffle(self, val: bool):
        return False

    @override
    def is_mute(self) -> bool:
        return False

    @override
    def can_go_next(self) -> bool:
        return True

    @override
    def can_go_previous(self) -> bool:
        return True

    @override
    def can_play(self) -> bool:
        return True

    @override
    def can_pause(self) -> bool:
        return True

    @override
    def can_seek(self) -> bool:
        return True

    @override
    def can_control(self) -> bool:
        return True

    @override
    def get_current_track(self) -> Track:
        song_data = self.player.list_of_downloaded_songs[self.player.current_song_index]
        title = song_data.title
        artist = [song_data.get_formatted_artists()]
        album = song_data.album
        length = self.player.song_length
        track = Track(
            title=title,
            artists=artist,
            album=album,
            length=length,
            art_url=f"https://i.ytimg.com/vi/{song_data.videoId}/maxresdefault.jpg",
        )

        return track

    @override
    def metadata(self) -> MetadataObj:
        song_data = self.player.list_of_downloaded_songs[self.player.current_song_index]
        title = song_data.title
        artist = [song_data.get_formatted_artists()]
        album = song_data.album
        length = self.player.song_length * 1000000
        metadata = MetadataObj(
            album=album,
            title=title,
            artists=artist,
            length=length,
            auto_rating=0,
            art_url=f"https://i.ytimg.com/vi/{song_data.videoId}/maxresdefault.jpg",
            url=f"https://www.youtube.com/watch?v={song_data.videoId}",
        )

        return metadata



import time
from unittest.mock import MagicMock
import pytest
from main import PyMusicTermPlayer, PyMusicTerm
from setting import SettingManager


@pytest.fixture
def app() -> PyMusicTermPlayer:
    setting = SettingManager()
    setting.os = "win32"
    app = PyMusicTerm(setting)
    return app.player


def test_seek_error(app: PyMusicTermPlayer):
    with pytest.raises(TypeError):
        app.seek("10")


def test_seek_forward(app: PyMusicTermPlayer):
    app.play_from_list(0)
    previous = int(
        app.music_player.position
    )  # int is used to round off the float because sometimes the float is not 0 but 0.003
    app.seek(10)
    assert int(app.music_player.position) == previous + 10


def test_seek_back(app: PyMusicTermPlayer):
    app.play_from_list(0)
    app.seek(10)
    previous = int(
        app.music_player.position
    )  # int is used to round off the float because sometimes the float is not 0 but 0.003
    app.seek(-10)
    assert int(app.music_player.position) == previous - 10


def test_loop_at_end(app: PyMusicTermPlayer):
    app.loop_at_end()
    assert app.setting.loop == app.music_player.loop_at_end


def test_check_if_song_ended_loop_at_end_true(app: PyMusicTermPlayer):
    app.play_from_list(0)
    app.music_player.loop_at_end = True
    if app.music_player.loop_at_end:
        app.check_if_song_ended = MagicMock()
        app.next = MagicMock()
        app.next.assert_not_called()


def test_check_if_song_ended_loop_at_end_false(app: PyMusicTermPlayer):
    app.play_from_list(0)
    app.music_player.loop_at_end = False
    app.music_player.position = app.music_player.song_length
    app.check_if_song_ended = MagicMock()
    app.next = MagicMock()
    app.check_if_song_ended()
    app.next.assert_called_once()



from pathlib import Path

import music_tag
from api.ytmusic import YTMusic
from api.protocols import SongData
from api.music_player import MusicPlayer
from api.downloader import Downloader
from player.media_control import MediaControl
from player.util import format_time
from setting import SettingManager, fetch_files_from_folder
from random import shuffle


class PyMusicTermPlayer:
    def __init__(
        self,
        setting: SettingManager,
        media_control: MediaControl,
    ) -> None:
        self.media_control = media_control
        self.setting = setting
        self.music_player = MusicPlayer(self.setting.volume)
        self.ytm = YTMusic()
        self.downloader = Downloader(self.setting.music_dir)
        self.list_of_downloaded_songs: list[SongData] = self.get_downloaded_songs()
        self.dict_of_song_result: dict[str, SongData] = {}
        self.current_song_index = 0
        self.current_song: SongData | None = None

    def get_downloaded_songs(self) -> list[SongData]:
        """Get the list of downloaded songs

        Returns:
            list[Song]: the list of downloaded songs
        """
        songs = fetch_files_from_folder(self.setting.music_dir, "mp3")
        list_of_songs = []
        for song in songs:
            song_metadata = music_tag.load_file(song)
            artist = song_metadata["artist"]
            list_of_songs.append(
                SongData(
                    title=str(song_metadata["title"]),
                    artist=artist.values,
                    duration=format_time(float(str(song_metadata["#length"]))),
                    videoId=Path(song).stem,
                    thumbnail=song_metadata["artwork"].first.thumbnail([128, 128]),
                    album=str(song_metadata["album"]),
                    path=Path(song),
                )
            )
        return list_of_songs

    def query(self, query: str, filter: str) -> list[SongData]:
        """Query the YTMusic API for a song

        Args:
            query (str): the query to search for
            filter (str): the filter to use

        Returns:
            list[SearchSongResult | SearchVideoResult]: the list of the results found
        """

        result = self.ytm.search(query, filter)

        self.dict_of_song_result.clear()
        for song in result:
            self.dict_of_song_result[song.videoId] = song
        return result

    def play_from_ytb(self, video_id: str) -> None:
        """Play a song from the YTMusic API, it will download the song first then play it

        Args:
            video_id (int): the video id of the song to play
        """
        song = self.dict_of_song_result[video_id]
        path = self.downloader.download(song)
        if path is None:
            return
        # BUG: dont re-fetch the songs but add it to the list
        self.list_of_downloaded_songs = self.get_downloaded_songs()
        self.music_player.load_song(str(path))
        self.music_player.play_song()
        self.media_control.on_playback()

    def play_from_list(self, id: int) -> None:
        """Play a song from the list of downloaded songs

        Args:
            id (int): the index of the song to play

        Raises:
            TypeError: If id is not an integer
        """
        if not isinstance(id, int):
            raise TypeError(f"id must be an integer, not {type(id)}")
        self.current_song_index = id
        self.current_song = self.list_of_downloaded_songs[id]
        self.music_player.load_song(self.list_of_downloaded_songs[id].path)
        self.music_player.play_song()
        self.media_control.on_playback()

    def previous(self) -> int:
        """Play the previous song"""
        if self.current_song_index == 0:
            self.current_song_index = len(self.list_of_downloaded_songs) - 1
        else:
            self.current_song_index -= 1
        self.music_player.load_song(
            self.list_of_downloaded_songs[self.current_song_index].path
        )
        self.music_player.play_song()
        self.media_control.on_playback()
        return self.current_song_index

    def next(self) -> int:
        """Play the next song"""
        if self.current_song_index == len(self.list_of_downloaded_songs) - 1:
            self.current_song_index = 0
        else:
            self.current_song_index += 1
        self.current_song = self.list_of_downloaded_songs[self.current_song_index]
        self.music_player.load_song(
            self.list_of_downloaded_songs[self.current_song_index].path
        )
        self.music_player.play_song()
        self.media_control.on_playback()
        return self.current_song_index

    def seek(self, seconds: float | int = 10) -> None:
        """Seek forward or backward

        Args:
            time (float, optional): The time to seek in seconds. Defaults to 10.

        Raises:
            TypeError: If seconds is not an integer or a float
        """
        if not isinstance(seconds, (int, float)):
            raise TypeError(
                f"Seconds must be an integer or a float, not {type(seconds)}"
            )
        self.music_player.position += seconds

    def seek_to(self, seconds: float | int) -> None:
        """Seek to a specific time

        Args:
            seconds (float | int): The time to seek in seconds

        Raises:
            TypeError: If seconds is not an integer or a float
        """
        if not isinstance(seconds, (int, float)):
            raise TypeError(
                f"Seconds must be an integer or a float, not {type(seconds)}"
            )
        self.music_player.position = seconds

    def suffle(self) -> None:
        """Shuffle the list of downloaded songs"""
        shuffle(self.list_of_downloaded_songs)
        if self.current_song is not None:
            self.current_song_index = self.list_of_downloaded_songs.index(
                self.current_song
            )
        else:
            self.current_song_index = 0
            self.current_song = self.list_of_downloaded_songs[self.current_song_index]

    def loop_at_end(self) -> bool:
        """Loop at the end"""
        self.music_player.loop_at_end = not self.music_player.loop_at_end
        self.setting.loop = self.music_player.loop_at_end
        return self.music_player.loop_at_end

    def check_if_song_ended(self) -> None:
        """Check if the song has ended and play the next song or loop it"""
        if self.music_player.loop_at_end:
            return
        if self.music_player.position == 0 and not self.music_player.playing:
            self.next()

    def stop(self) -> None:
        self.music_player.unload_song()

    @property
    def song_length(self) -> float:
        """Get the song length"""
        return self.music_player.song_length

    @property
    def position(self) -> float:
        """Get the current position"""
        return self.music_player.position

    def volume(self, value: float) -> None:
        """Get the volume up"""
        self.music_player.volume += value
        self.media_control.on_volume()
        self.setting.volume = self.music_player.volume

    @property
    def playing(self) -> bool:
        """Get the playing status"""
        return self.music_player.playing

    def pause_song(self) -> None:
        """Pause the song"""
        self.music_player.pause_song()
        self.media_control.on_playpause()

    def resume_song(self) -> None:
        """Resume the song"""
        self.music_player.resume_song()
        self.media_control.on_playpause()



import pytest
from unittest.mock import patch
from api.notification_manager import NotificationManager


def test_send_notification_success():
    with patch("notifypy.Notify.send") as mock_notify:
        manager = NotificationManager()
        manager.send_notification("Test Title", "Test Message")
        mock_notify.assert_called_once()


def test_send_notification_invalid_title():
    manager = NotificationManager()
    with pytest.raises(TypeError):
        manager.send_notification(123, "Test Message")


def test_send_notification_invalid_message():
    manager = NotificationManager()
    with pytest.raises(TypeError):
        manager.send_notification("Test Title", 123)



from datetime import timedelta


def format_time(seconds: int | float) -> str:
    """Format the time to a string (ex: 90s -> 01:30)

    Args:
        seconds (int | float): The time in secondes

    Raises:
        TypeError: If the type of the seconds is not int or float

    Returns:
        str: The string representation of the time
    """
    if not isinstance(seconds, (int, float)):
        raise TypeError(
            f"Invalid type for time for {type(seconds)}, please use int or float"
        )
    if isinstance(seconds, float):
        seconds = int(seconds)
    if seconds < 0:
        seconds = 0
    seconds = str(timedelta(seconds=seconds)).removeprefix("0:").removeprefix("0")
    return seconds
